/*->c.fcet */

#include <stdio.h>

#define FCET_DECODER
#define FCET_ENCODER

#include "h.fcet"


#define POLYNOMIAL      (0xedb88320)    /* Polynomial used by CRC checksum */


typedef int boolean;
#define FALSE   (0)
#define TRUE    (1)


/* Level 1 codes */

#define OSEP    (257)   /* Section separator */
#define OEOF    (258)   /* End of file marker */
#define OERROR  (259)   /* Error during decoding process */


/* Level 2 codes */

#define CRLE0   (0)     /* RLE introducer for zeros */
#define CRLE    (1)     /* RLE introducer for non-zeros */
#define C0START (2)     /* Start codde for byte data */
#define CSEP    (259)   /* Section separator */
#define CEOF    (260)   /* End of file marker */
#define CERROR  (261)   /* Error during decoding process */


/* Level 3 codes */

#define H0      (81)    /* Introducers for 8-blocks */
#define H1      (82)
#define H2      (83)

#define HRLE0   (88)    /* Introducers for variable length blocks */
#define HRLE1   (89)
#define HRLE2   (90)
#define HREND   (91)    /* End of variable length block marker */

#define SPLITFACT (87)  /* Range of loval values */
#define RANGE_START (32) /* Start of displayable character range */


/* Inter-function static variables */

static int      S_eorsum, S_rolsum;     /* Simple checksum workspace */
static uint32   S_crc;                  /* 32 bit CRC workspace */
static uint32   S_crc32tab[256];        /* CRC generation table */
static char S_marker[] = "---FCET---";  /* FCET recognition marker */


/*******************************  fcet_init()  ********************************
 * This function generates the 256 entry table used to perform CRC
 * calculations 8 bits at a time.
 *
 * This table need only be set up once and can be used to decode and encode a
 * number of files.
 */

void fcet_init()

{
        register int i, bit;
        register uint32 crcsum;

        for (i = 0; i <= 255; i++)
                S_crc32tab[i] = 0;

        crcsum = POLYNOMIAL;
        for (bit = 128; bit; bit >>= 1) { /* Do each bit in turn hi - lo */
                for (i = 0; i <= 0xff; i++) {
                        if (i & bit)
                                S_crc32tab[i] ^= crcsum;
                }

                if (crcsum & 1) {
                        crcsum >>= 1;
                        crcsum ^= POLYNOMIAL;
                } else
                        crcsum >>= 1;
        }
}


/******************************************************************************
****************************  DECODER DEFINITIONS  ****************************
******************************************************************************/

#ifdef FCET_DECODER


/* Internal (static) function definitions */

static int      test_checksum(void);
static int      readlong(int32 *);
static int      dec_level2(void);
static int      dec_level3(void);


/* Inter-function static variables */

static FILE     *S_input_fp;            /* Input file pointer */
static int S_prevch, S_rlecount;        /* Level 2 status */
static enum {                           /* Level 3 status */
        L3_WAITING, L3_INBLOCK, L3_INBLOCK8, L3_INRLE
} S_dl3state;
static int S_markcount;         /* Number of FCET marker characters matched */
static int S_markch;            /* The character which didn't match */



/******************************  fcetdec_hdr()  *******************************
 * Decodes an FCET header. The function returns a success code and the header
 * data is returned through the parameters. A return parameter is ignored if
 * the pointer is NULL.
 */

int fcetdec_hdr(input_fp, faptr)
FILE    *input_fp;      /* File stream pointer for input file */
struct Fcetargs *faptr; /* Pointer to return file arguments */

{
        register int i, ch, arglist;
        register char *filename;

        S_input_fp = input_fp;

/* Check for FCET marker string */

        for (i = 0; i < 10; i++) {
                if ((ch = getc(input_fp)) != S_marker[i]) {
                        S_markcount = i;
                        S_markch = ch;
                        return(FRC_NOTFCET);
                }
        }

        S_markcount = -1;       /* Recognised FCET format */

/* Initialise checksum variables for header */

        S_crc = 0xffffffff;
        S_eorsum = 0;
        S_rolsum = 0;

/* Initialise decoder state variables */

        S_prevch = 0;
        S_rlecount = 0;
        S_dl3state = L3_WAITING;

/* Read header data */

        if ((arglist = dec_level2()) > 255      /* Read arglist flags */
        || readlong(&faptr->length))            /* Read file length */
                return(FRC_FORMAT);
        faptr->arglist = arglist;

        for (i = 0; i < 8; i++) {               /* Read argument list */
                if (arglist & 1) {
                        if (readlong(&faptr->arg[i]))
                                return(FRC_FORMAT);
                }

                arglist >>= 1;
        }

        filename = faptr->filename;     /* Read file name */
        i = 0;
        while ((ch = dec_level2()) <= 255) {
                if (i == 12)
                        return(FRC_FORMAT);

                filename[i++] = ch;
        }
        filename[i] = '\0';

/* Check for end of header separator */

        if (ch != OSEP)
                return(FRC_FORMAT);

/* Read and test header checksum values */

        if (test_checksum())
                return(FRC_CHECKSUM);

/* Check for end of checksum separator */

        if ((ch = dec_level2()) != OSEP)
                return(FRC_FORMAT);

        return(FRC_OK);
}


/******************************  fcetdec_data()  ******************************
 * Reads an FCET encoded file from input_fp and decodes it.
 *
 * outname provides the output filename or if it null then the filename given
 * in the header is used.
 */

int fcetdec_data(output_fp, lengthptr)
register FILE *output_fp;       /* File where decoded output data is sent */
int32   *lengthptr;             /* Return for actual data length */

{
        register int ch;

/* Just copy file if not FCET format */

        if (S_markcount >= 0) {
                register int i;

                for (i = 0; i < S_markcount; i++) /* Output chs already read */
                        putc(S_marker[i], output_fp);

                for (ch = S_markch; ch != EOF; ch = getc(S_input_fp))
                        putc(ch, output_fp);

                return(FRC_OK);
        }

/* Initialise checksum variables for data */

        S_crc = 0xffffffff;
        S_eorsum = 0;
        S_rolsum = 0;

/* Read and output data */

        {
                register int32 length;

                length = 0;
                for (;;) {
                        if ((ch = dec_level2()) <= 255) {
                                putc(ch, output_fp);
                                length++;
                        } else if (ch == OSEP)
                                break;
                        else
                                return(FRC_FORMAT);
                }

                if (lengthptr)
                        *lengthptr = length;
        }

/* Read and test data checksum values */

        if (test_checksum())
                return(FRC_CHECKSUM);

/* Check for end of file marker */

        if ((ch = dec_level2()) != OEOF)
                return(FRC_FORMAT);

        return(FRC_OK);
}


/*****************************  test_checksum()  ******************************
 * Checks to see if the checksum is OK.
 *
 * Returns TRUE if there is an error.
 */

static int test_checksum()

{
        if (readlong(0) || S_crc ||
        dec_level2() > 255 || S_eorsum ||
        dec_level2() > 255 || (S_rolsum & 0xff))
                return(TRUE);

        return(FALSE);
}


/********************************  readlong()  ********************************
 * Read a long integer from dec_level2().
 *
 * The value is returned through longret.
 *
 * The function returns TRUE if a format error is encountered.
 */

static int readlong(longret)
int32   *longret;

{
        register int i, ch;
        register int32 longval;

        longval = 0;
        for (i = 0; i < 32; i += 8) {
                if ((ch = dec_level2()) > 255)
                        return(TRUE);

                longval |= (int32) ch << i;
        }

        if (longret)
                *longret = longval;

        return(FALSE);
}


/*******************************  dec_level2()  *******************************
 * This performs level 2 decoding:
 *      a) Decoding of Run-Length-Encoding sequences
 *      b) Exclusive-OR with previous character
 *      c) Checksum calculation
 */

static int dec_level2()

{
        register int ch;
        static int S_rlech;     /* Character which is RLE'd */
        static int S_lastch;    /* Last non-zero output character */

/* Perform RLE decoding */

        if (S_rlecount) {
                S_rlecount--;           /* In RLE - reduce RLE count and */
                ch = S_rlech;           /* return current RLE character */
        } else if ((ch = dec_level3()) >= C0START) {
                if (ch -= C0START)      /* Pass ordinary character + ERROR */
                        S_lastch = ch;
        } else {                        /* Decode RLE sequence */
                switch (ch) {
                case CRLE0:
                        ch = 0;
                        break;
                case CRLE:
                        ch = S_lastch;
                        break;
                default:
                        return(OERROR);
                }

                S_rlech = ch;           /* Set RLE char and initial count */
                S_rlecount = dec_level3() + 2;  /* Plus this one makes 3 */

                if (S_rlecount == CERROR + 2)
                        return(OERROR);
        }

/* Exclusive-OR with previous character and update checksum */

        if (ch <= 255) {
                ch ^= S_prevch;         /* EOR character with previous */
                S_prevch = ch;

                S_crc = S_crc32tab[ch ^ (S_crc & 0xff)] ^ (S_crc >> 8);/*CRC*/

                S_eorsum ^= ch;         /* EOR checksum */

                if ((S_rolsum <<= 1) & 0x100)   /* Rotation EOR checksum */
                        S_rolsum++;
                S_rolsum ^= ch;
                S_rolsum++;
        }

        return(ch);
}


/*******************************  dec_level3()  *******************************
 * This performs level 3 decoding:
 *      a) The display range offset (RANGE_START) is removed 
 *      b) Level 3 block types are decoded
 */

static int dec_level3()

{
        register int ch;

        static int S_charcount, S_hivals;
        static int S_times_sf[] = { 0, SPLITFACT, 2*SPLITFACT };

        for (;;) {      /* Loop until we have a value to return */

/* Get input character */

                if ((ch = getc(S_input_fp)) == EOF)
                        return(CERROR);

/* Remove display offset */

                ch -= RANGE_START;

/* State driven level 3 decoding */

                switch (S_dl3state) {
                case L3_WAITING:        /* Looking for block introducer */
                        if (ch < 81) {          /* Basic 4-5 block */
                                S_hivals = ch;
                                S_charcount = 4;
                                S_dl3state = L3_INBLOCK;
                        } else if (ch >= H0 && ch <= H2) { /* 8 char block */
                                S_hivals = S_times_sf[ch - H0];
                                S_charcount = 8;
                                S_dl3state = L3_INBLOCK8;
                        } else if (ch >= HRLE0 && ch <= HRLE2) { /* Variable */
                                S_hivals = S_times_sf[ch - HRLE0]; /* block */
                                S_dl3state = L3_INRLE;
                        } else
                                return(CERROR);

                        continue;
                case L3_INBLOCK:        /* Within a 4-5 block */
                        if (ch >= SPLITFACT)
                                return(CERROR);

                        ch += S_times_sf[S_hivals % 3];
                        S_hivals /= 3;

                        if (!--S_charcount)
                                S_dl3state = L3_WAITING;

                        break;
                case L3_INBLOCK8:       /* Within an 8 character block */
                        if (ch >= SPLITFACT)
                                return(CERROR);

                        ch += S_hivals;

                        if (!--S_charcount)
                                S_dl3state = L3_WAITING;

                        break;
                default: /* L3_INRLE        Within a variable length blk */
                        if (ch >= SPLITFACT) {
                                if (ch >= HRLE0 && ch <= HRLE2)
                                        S_hivals = S_times_sf[ch - HRLE0];
                                else if (ch == HREND)
                                        S_dl3state = L3_WAITING;
                                else
                                        return(CERROR);

                                continue;
                        }

                        ch += S_hivals;
                        break;
                }

                return(ch);
        }
}

#endif

/******************************************************************************
****************************  ENCODER DEFINITIONS  ****************************
******************************************************************************/

#ifdef FCET_ENCODER


/* Local static functions */

static void     write_csum(void);
static void     writelong(int32);
static void     level3(int);


/* General static variables */

static boolean  S_finalch;              /* Final character flag */


/* Level 2 variables initialised in fcetenc_hdr() */

static long     S_l2runlen;             /* RLE run length */
static int      S_l2runch;              /* RLE run character */
static int      S_l2lastoutch;          /* Last non-zero O/P char. */
static int      S_l2prevch;             /* Previous I/P character */


/* Level 3 variables initialised in fcetenc_hdr() */

static enum {                           /* Level 3 state */
        L3_BLOCK1, L3_BLOCK2, L3_RUN
} S_el3state;
static int      S_l3runlen;             /* Run length */
static boolean  S_l3termflag;           /* RLE terminator flag */


/******************************  fcetenc_hdr()  *******************************
 * Write out an FCET header.
 */

void fcetenc_hdr(faptr)
register struct Fcetargs *faptr;

{
        register int i, ch, arglist;
        register char *filename;

/* Write FCET marker */

        for (i = 0; ch = S_marker[i]; i++)
                fcetenc_out(ch);

/* Initialise checksum calculation variables for header */

        S_crc = 0xffffffff;
        S_eorsum = 0;
        S_rolsum = 0;

/* Initialise encoder state variables so that encoder can be recalled */

        S_finalch = FALSE;

        S_l2runlen = 0;
        S_l2runch = -1;
        S_l2lastoutch = 0;
        S_l2prevch = 0;

        S_el3state = L3_BLOCK1;
        S_l3runlen = 0;
        S_l3termflag = FALSE;

/* Write out header data */

        fcetenc_l2(faptr->arglist);     /* Argument list bitmask */
        writelong(faptr->length);       /* File length */

        arglist = faptr->arglist;       /* Send argument list */
        for (i = 0; i < 8; i++) {
                if (arglist & 1)
                        writelong(faptr->arg[i]);

                arglist >>= 1;
        }

        filename = faptr->filename;
        while (ch = *filename++)        /* Filename */
                fcetenc_l2(ch);

/* Write out header checksum */

        write_csum();

        fcetenc_l2(OSEP);               /* Separator */

/* Initialise checksum calculation variables for data */

        S_crc = 0xffffffff;
        S_eorsum = 0;
        S_rolsum = 0;
}


/******************************  fcetenc_tail()  ******************************
 * Feeds characters from the source into the converter
 * The last character sent should be EOF
 */

void fcetenc_tail()

{

/* Write out data checksum */

        write_csum();                   /* Data checksum */

        fcetenc_l2(OEOF);               /* Terminator */
}


/*******************************  write_csum()  *******************************
 * Write a checksum block
 */

static void write_csum()

{
        fcetenc_l2(OSEP);               /* Separator */

        writelong(S_crc);               /* CRC */
        fcetenc_l2(S_eorsum);           /* EOR */
        if ((S_rolsum <<= 1) & 0x100)   /* Rotate/EOR */
                S_rolsum++;
        fcetenc_l2((S_rolsum & 0xff) ^ 0xff);
}


/*******************************  writelong()  ********************************
 * Output a long integer low byte first
 */

static void writelong(longval)
register int32 longval;

{
        register int i;

        for (i = 0; i < 4; i++) {
                fcetenc_l2(longval & 0xff);
                longval >>= 8;
        }
}


/*******************************  fcetenc_l2()  *******************************
 * Level 2 FCET encoding. 
 */

void fcetenc_l2(ch)
register int ch;                        /* Input character */

{
        register int tempch;            /* Character temporary store */
        int     sublength;

        
/* Perform Exclusive-OR and checksum calculations */

        if (ch < 256) {
                S_crc = S_crc32tab[ch ^ (S_crc & 0xff)] ^ (S_crc >> 8);

                S_eorsum ^= ch;

                if ((S_rolsum <<= 1) & 0x100)
                        S_rolsum++;
                S_rolsum ^= ch;
                S_rolsum++;

                tempch = S_l2prevch;    /* EOR character with previous */
                S_l2prevch = ch;
                ch ^= tempch;
        }

        ch += C0START;          /* Adjust to C0START */

/* Perform character run length encoding */

        if (ch != S_l2runch) {
                if (S_l2runlen) {
                        tempch = S_l2runch;

                        if (tempch != C0START && tempch != S_l2lastoutch) {
                                level3(tempch);
                                S_l2runlen--;
                                S_l2lastoutch = tempch;
                        }

                        while (S_l2runlen) {
                                if (S_l2runlen < 3) {
                                        do {    /* Too small for RLE */
                                                level3(tempch);
                                        } while (--S_l2runlen);

                                        if (tempch != C0START)
                                                S_l2lastoutch = tempch;

                                        break;
                                }

                                S_l2runlen -= 3;
                                sublength = (S_l2runlen < 260)
                                                ? S_l2runlen : 260;

                                level3((tempch == C0START) ? CRLE0 : CRLE);
                                level3(sublength);

                                S_l2runlen -= sublength;
                        }
                } /* S_l2runlen must be 0 at this point */

                S_l2runch = ch;         /* Set up for new run character */
        }

        S_l2runlen++;

        if (ch == CEOF) {               /* Don't delay CEOF */
                S_finalch = TRUE;
                level3(ch);
        }
}


/*********************************  level3()  *********************************
 * This performs level 3 encoding:
 *      a) Level 3 block types are encoded
 *      b) The display offset (RANGE_START) is added
 */

static void level3(value)
int     value;

{
        register int i, block1size, hienc;
        register int loval, hival;
        static int S_currhi;
        static int S_lovals[12], S_hivals[12];  /* Look-ahead buffer */

/* Split input value up into loval,hival pair */

        loval = value % SPLITFACT;
        hival = value / SPLITFACT;

        switch (S_el3state) {
        case L3_BLOCK1:
                S_lovals[S_l3runlen] = loval;
                S_hivals[S_l3runlen++] = hival;

                if (S_finalch)
                        break;

                if (S_l3runlen == 4) {
                        if (hival != S_hivals[0] ||
                            hival != S_hivals[1] ||
                            hival != S_hivals[2])
                                break;

                        S_currhi = hival;
                        S_el3state = L3_BLOCK2;
                }

                return;
        case L3_BLOCK2:
                S_lovals[S_l3runlen] = loval;
                S_hivals[S_l3runlen++] = hival;

                if (hival != S_currhi)
                        break;
                
                if (S_l3runlen == 12 || S_finalch) {
                        fcetenc_out(hival + HRLE0 + RANGE_START);
                        for (i = 0; i < S_l3runlen; i++)
                                fcetenc_out(S_lovals[i] + RANGE_START);

                        S_el3state = L3_RUN;
                }

                return;
        default: /* L3_RUN */
                if (hival == S_currhi)
                        fcetenc_out(loval + RANGE_START);
                else {
                        S_lovals[0] = loval;
                        S_hivals[0] = hival;
                        S_l3termflag = TRUE;
                        S_el3state = L3_BLOCK1;
                        S_l3runlen = 1;

                        if (S_finalch)
                                break;
                }

                return;
        }

/* Write out buffer as a 4 and/or 8 block */

        if (S_l3termflag) {             /* RLE terminator */
                fcetenc_out(HREND + RANGE_START);
                S_l3termflag = FALSE;
        }

        do {
                if ((block1size = S_l3runlen) > 4) {
                        if (block1size > 8) {   /* Write out 8 block */
                                fcetenc_out(S_currhi + H0 + RANGE_START);
                                for (i = 0; i < 8; i++)
                                        fcetenc_out(S_lovals[i] + RANGE_START);

                                do {
                                        S_lovals[i - 8] = S_lovals[i];
                                        S_hivals[i - 8] = S_hivals[i];
                                } while (++i < block1size);

                                S_l3runlen -= 8;
                                continue;
                        }

                        block1size = 4;
                }

                hienc = 0;              /* Encode hivals */
                i = block1size;
                while (i--)
                        hienc += (hienc << 1) + S_hivals[i];

                fcetenc_out(hienc + RANGE_START); /* Write hivals and lovals */
                for (i = 0; i < block1size; i++)
                        fcetenc_out(S_lovals[i] + RANGE_START);

                for (i = 4; i < S_l3runlen; i++) {
                        S_lovals[i - 4] = S_lovals[i];
                        S_hivals[i - 4] = S_hivals[i];
                }

                S_l3runlen -= block1size;
        } while (S_l3runlen == 4 || (S_finalch && S_l3runlen));

        S_el3state = L3_BLOCK1;
}

#endif
